@@ -30,10 +30,6 @@
 
 class URI;
 
-// FIXME: kill this export on the next ABI break - strongly doubt its in use anyway
-// outside of the apt itself, its really a internal interface
-APT_DEPRECATED int parsenetrc (char *host, char *login, char *password, char *filename);
-
 void maybe_add_auth (URI &Uri, std::string NetRCFile);
 bool IsAuthorized(pkgCache::PkgFileIterator const I) APT_HIDDEN;
 #endif
